from typing import Callable, Dict, Optional, Tuple, Union

import numpy as np
import torch
from pyro.infer import HMC, MCMC
from torch.distributions import Normal  # noqa: F401
from torch.distributions import Categorical
from torch.distributions import Distribution as torchDist
from torch.nn import functional as F
from tqdm import trange

from .gan_distribution import Distribution


class MCMCRegistry:
    registry: Dict = {}

    @classmethod
    def register(cls, name: Optional[str] = None) -> Callable:
        def inner_wrapper(function: Callable) -> Callable:
            if name is None:
                name_ = function.__name__
            else:
                name_ = name
            cls.registry[name_] = function
            return function

        return inner_wrapper

    @classmethod
    def __call__(cls, name: str, *args, **kwargs) -> Tuple[torch.FloatTensor, Dict]:
        exec = cls.registry[name]
        # kwargs["proposal"] = kwargs.get("proposal", Normal(0, 1))
        kwargs["project"] = kwargs.get("project", lambda _: _)
        chains, meta = exec(*args, **kwargs)
        return chains, meta


@MCMCRegistry.register()
def ula(
    start: torch.FloatTensor,
    target: Union[Distribution, torchDist],
    proposal: Union[Distribution, torchDist],
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    step_size: float,
    verbose: bool = False,
    meta: Optional[Dict] = None,
    keep_graph: bool = False,
    natural_grad: bool = False,  # True,
    tikhonov_reg: float = 0.1,
) -> Tuple[torch.FloatTensor, Dict]:
    """
    Unadjusted Langevin Algorithm

    Args:
        start - strating points of shape [n_chains, dim]
        target - target distribution instance with method "log_prob"
        n_samples - number of last samples from each chain to return
        burn_in - number of first samples from each chain to throw away
        step_size - step size for drift term
        verbose - whether to show iterations' bar

    Returns:
        tensor of chains with shape [n_samples, n_chains, dim]
    """
    chains = []
    meta = meta or dict()

    point = start.clone()
    point.requires_grad_(True)
    point.grad = None

    pbar = trange if verbose else range
    for step_id in pbar(n_samples + burn_in):
        logp = target.log_prob(point)
        grad = torch.autograd.grad(
            logp.sum(), point, create_graph=keep_graph, retain_graph=keep_graph
        )[0]
        if natural_grad:
            step_size_ = step_size / (
                tikhonov_reg + torch.norm(grad, dim=-1)[:, None].detach()
            )
        else:
            step_size_ = step_size

        noise = torch.randn_like(point, dtype=torch.float).to(point.device)
        noise_scale = (2.0 * step_size_) ** 0.5

        point = point + step_size_ * grad + noise_scale * noise
        # point = point / torch.clamp(torch.norm(point, dim=-1), min=1)[:, None] * point.shape[-1]**.5
        point = project(point)

        if not keep_graph:
            point = point.detach().requires_grad_()
        if step_id >= burn_in:
            chains.append(point.cpu().clone())
    chains = torch.stack(chains, 0)
    meta["mask"] = torch.zeros(point.shape[0], dtype=torch.bool)

    return chains, meta


@MCMCRegistry.register()
# @torch.no_grad()
def isir_step(
    start: torch.FloatTensor,
    target,
    proposal: Union[Distribution, torchDist],
    *,
    n_particles: int,
    logp_x=None,
    logq_x=None,
) -> Tuple:
    point = start.clone()
    # logp_x = target.log_prob(point) if logp_x is None else logp_x
    logq_x = proposal.log_prob(point) if logq_x is None else logq_x

    particles = proposal.sample((point.shape[0], n_particles - 1))
    log_qs = torch.cat([logq_x[:, None], proposal.log_prob(particles)], 1)
    # log_ps = torch.cat([logp_x[:, None], target.log_prob(particles)], 1)
    particles = torch.cat([point[:, None, :], particles], 1)
    log_ps = target.log_prob(particles)

    log_weights = log_ps - log_qs
    indices = Categorical(logits=log_weights).sample()

    x = particles[np.arange(point.shape[0]), indices]

    return x, particles, log_ps, log_qs, indices


@MCMCRegistry.register()
@torch.no_grad()
def isir(
    start: torch.FloatTensor,
    target,
    proposal: Union[Distribution, torchDist],
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    n_particles: int,
    verbose: bool = False,
    meta: Optional[Dict] = None,
    **kwargs,
) -> Tuple[torch.FloatTensor, Dict]:
    """
    Iterated Sampling Importance Resampling

    Args:
        start - strating points of shape [n_chains x dim]
        target - target distribution instance with method "log_prob"
        proposal - proposal distribution instance with methods "log_prob" and "sample"
        n_samples - number of last samples from each chain to return
        burn_in - number of first samples from each chain to throw away
        n_particles - number of particles including one from previous step
        verbose - whether to show iterations' bar

    Returns:
        tensor of chains with shape [n_samples, n_chains, dim], meta
    """
    chains = []
    point = start.clone()
    # point = project(start)

    meta = meta or dict()
    meta["sir_accept"] = meta.get("sir_accept", [])
    meta["logp"] = meta.get("logp", target.log_prob(point))
    logp_x = meta["logp"]
    logq_x = proposal.log_prob(point)

    pbar = trange if verbose else range
    for step_id in pbar(n_samples + burn_in):
        point, _, log_ps, log_qs, indices = isir_step(
            point,
            target,
            proposal,
            n_particles=n_particles,
            logp_x=logp_x,
            logq_x=logq_x,
        )
        logp_x = log_ps[np.arange(point.shape[0]), indices]
        logq_x = log_qs[np.arange(point.shape[0]), indices]
        meta["sir_accept"].append((indices != 0).float().mean().item())

        if step_id >= burn_in:
            chains.append(point.cpu().clone())
    chains = torch.stack(chains, 0)

    meta["logp"] = logp_x
    meta["mask"] = F.one_hot(indices, num_classes=n_particles).to(bool).detach().cpu()

    return chains, meta


def heuristics_step_size(
    mean_acceptance: float,
    target_acceptance: float,
    step_size: float,
    factor: float = 1.05,
    tol: float = 0.03,
):
    if mean_acceptance - target_acceptance > tol:
        return step_size * factor
    if target_acceptance - mean_acceptance > tol:
        return step_size / factor
    return step_size


@MCMCRegistry.register()
def mala(
    start: torch.FloatTensor,
    target: Union[Distribution, torchDist],
    proposal: Union[Distribution, torchDist],
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    step_size: float,
    verbose: bool = False,
    target_acceptance=None,
    meta: Optional[Dict] = None,
    keep_graph: bool = False,
) -> Tuple[torch.FloatTensor, Dict]:
    """
    Metropolis-Adjusted Langevin Algorithm with Normal proposal

    Args:
        start - strating points of shape [n_chains x dim]
        target - target distribution instance with method "log_prob"
        step_size - step size for drift term
        verbose - whether show iterations' bar

    Returns:
        sequence of slices per each iteration, meta
    """
    if n_samples + burn_in <= 0:
        raise ValueError("Number of steps might be positive")

    chains = []
    point = start.clone()
    point.requires_grad_()
    point.grad = None

    device = point.device
    proposal = torch.distributions.MultivariateNormal(
        torch.zeros(point.shape[-1], device=device),
        torch.eye(point.shape[-1], device=device),
    )

    meta = meta or dict()
    meta["mh_accept"] = meta.get("mh_accept", [])
    meta["step_size"] = meta.get("step_size", [])

    meta["logp"] = logp_x = target.log_prob(
        point
    )  # meta.get("logp", target.log_prob(point))
    if "grad" not in meta:
        if keep_graph:
            grad_x = torch.autograd.grad(
                meta["logp"].sum(),
                point,
                create_graph=keep_graph,
                retain_graph=keep_graph,
            )[0]
        else:
            grad_x = torch.autograd.grad(logp_x.sum(), point)[0].detach()
        meta["grad"] = grad_x
    else:
        grad_x = meta["grad"]

    pbar = trange if verbose else range
    for step_id in pbar(n_samples + burn_in):
        noise = proposal.sample(point.shape[:-1])
        proposal_point = point + step_size * grad_x + noise * (2 * step_size) ** 0.5
        proposal_point = project(proposal_point)
        if not keep_graph:
            proposal_point = proposal_point.detach().requires_grad_()

        logp_y = target.log_prob(proposal_point)
        grad_y = torch.autograd.grad(
            logp_y.sum(),
            proposal_point,
            create_graph=keep_graph,
            retain_graph=keep_graph,
        )[
            0
        ]  # .detach()

        log_qyx = proposal.log_prob(noise)
        log_qxy = proposal.log_prob(
            (point - proposal_point - step_size * grad_y) / (2 * step_size) ** 0.5
        )

        accept_prob = torch.clamp((logp_y + log_qxy - logp_x - log_qyx).exp(), max=1)
        mask = torch.rand_like(accept_prob) < accept_prob
        mask = mask.detach()

        if keep_graph:
            mask_f = mask.float()
            point = point * (1 - mask_f)[:, None] + proposal_point * mask_f[:, None]
            logp_x = logp_x * (1 - mask_f) + logp_y * mask_f
            grad_x = grad_x * (1 - mask_f)[:, None] + grad_y * mask_f[:, None]
        else:
            with torch.no_grad():
                point[mask, :] = proposal_point[mask, :]
                logp_x[mask] = logp_y[mask]
                grad_x[mask] = grad_y[mask]

        meta["mh_accept"].append(mask.float().mean().item())
        if target_acceptance:
            step_size = heuristics_step_size(
                meta["mh_accept"][-1], target_acceptance, step_size
            )
        meta["step_size"].append(step_size)

        if not keep_graph:
            point = point.detach().requires_grad_()
        if step_id >= burn_in:
            chains.append(point.cpu().clone())
    chains = torch.stack(chains, 0)

    meta["logp"] = logp_x
    meta["grad"] = grad_x
    meta["mask"] = mask.detach().cpu()

    return chains, meta


@MCMCRegistry.register()
def ex2mcmc(
    start: torch.FloatTensor,
    target: Union[Distribution, torchDist],
    proposal: Union[Distribution, torchDist],
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    step_size: float,
    n_particles: int,
    n_mala_steps: int = 1,
    target_acceptance: float = False,
    verbose: bool = False,
    meta: Optional[Dict] = None,
    keep_graph: bool = False,
):
    chains = []
    point = start.clone()
    point.requires_grad_(True)
    point.grad = None

    meta = meta or dict()
    meta["sir_accept"] = meta.get("sir_accept", [])
    meta["mh_accept"] = meta.get("mh_accept", [])
    meta["step_size"] = meta.get("step_size", [])

    pbar = trange if verbose else range
    meta["logp"] = meta.get("logp", target.log_prob(point))
    for step_id in pbar(n_samples + burn_in):
        # points, meta = isir(
        #     point, target, proposal, 1, 0, project, n_particles=n_particles, meta=meta
        # )
        # meta["grad"] = torch.autograd.grad(meta["logp"].sum(), points[-1])[
        #     0
        # ]  # .detach()
        logp_x = meta["logp"]
        logq_x = proposal.log_prob(point)

        point, proposals, log_ps, log_qs, indices = isir_step(
            point,
            target,
            proposal,
            n_particles=n_particles,
            logp_x=logp_x,
            logq_x=logq_x,
        )
        logp_x = log_ps[np.arange(point.shape[0]), indices]
        meta["sir_accept"].append((indices != 0).float().mean().item())
        point = point.detach().requires_grad_()
        meta["logp"] = logp_x
        meta["mask"] = (
            F.one_hot(indices, num_classes=n_particles).to(bool).detach().cpu()
        )

        points, meta = mala(
            point,  # s[-1],
            target,
            proposal,
            n_mala_steps,
            n_mala_steps - 1,
            project,
            step_size=step_size,
            target_acceptance=target_acceptance,
            meta=meta,
            keep_graph=keep_graph,
        )
        step_size = meta["step_size"][-1]
        point = points[-1]
        if not keep_graph:
            point = point.detach().requires_grad_()
        if step_id >= burn_in:
            chains.append(point.cpu().clone())

    chains = torch.stack(chains, 0)
    return chains, meta


@MCMCRegistry.register()
def hmc(
    start: torch.FloatTensor,
    target: Union[Distribution, torchDist],
    proposal: Union[Distribution, torchDist],
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    step_size: float,
    leapfrog_steps: int = 1,
    target_acceptance: float = False,
    adapt_step_size: bool = False,
    verbose: bool = False,
    meta: Optional[Dict] = None,
):
    meta = meta or dict()
    meta["step_size"] = meta.get("step_size", [])
    if "hmc_kernel" not in meta:
        kernel = HMC(
            potential_fn=lambda x: -target.log_prob(x["points"]),
            full_mass=False,
            step_size=step_size,
            num_steps=leapfrog_steps,
            adapt_step_size=adapt_step_size,
            target_accept_prob=target_acceptance,
        )
        meta["hmc_kernel"] = kernel

    x = start.clone()
    x.requires_grad_(False)
    x.grad = None

    init_params = {"points": x}
    mcmc = MCMC(
        kernel=meta["hmc_kernel"],
        num_samples=n_samples,
        initial_params=init_params,
        warmup_steps=burn_in,
    )
    mcmc.run()
    meta["step_size"].append(meta["hmc_kernel"].step_size)

    chains = mcmc.get_samples(group_by_chain=True)["points"]
    chains = chains.view(-1, *start.shape)

    return chains, meta


@MCMCRegistry.register()
def flex2mcmc(
    start: torch.FloatTensor,
    target,
    proposal,
    n_samples: int,
    burn_in: int,
    project: Callable,
    *,
    n_particles: int,
    step_size: float,
    n_mala_steps: int = 1,
    # add_pop_size_train: int = 200, #4096,
    forward_kl_weight: float = 1.0,
    backward_kl_weight: float = 1.0,
    entr_weight: float = 0.1,
    target_acceptance: float = False,
    verbose: bool = False,
    meta: Optional[Dict] = None,
    keep_graph: bool = False,
) -> Tuple[torch.FloatTensor, Dict]:
    """
    Ex2MCMC with Flow proposal

    Args:
        start - strating points of shape [n_chains x dim]
        target - target distribution instance with method "log_prob"
        proposal - instance of RealNVProposal
        n_samples - number of last samples from each chain to return
        burn_in - number of first samples from each chain to throw away
        step_size - step size for drift term
        n_particles - number of particles including one from previous step
        n_mala_steps - number of MALA steps after each SIR step
        verbose - whether to show iterations' bar

    Returns:
        tensor of chains with shape [n_samples, n_chains, dim],
          acceptance rates for each iteration
    """
    chains = []
    if meta:
        meta["sir_accept"] = meta.get("sir_accept", [])
        meta["forward_kl"] = meta.get("forward_kl", [])
        meta["backward_kl"] = meta.get("backward_kl", [])
    else:
        meta = dict(sir_accept=[], forward_kl=[], backward_kl=[])

    point = start.clone()
    point.requires_grad_(True)
    point.grad = None

    # hist_proposals = None

    meta["logp"] = target.log_prob(point)  #
    meta["logp"] = meta.get("logp", target.log_prob(point))

    pbar = trange(n_samples + burn_in) if verbose else range(n_samples + burn_in)
    for step_id in pbar:
        logp_x = meta["logp"]
        # logq_x = proposal.log_prob(point)

        point, _, log_ps, log_qs, indices = isir_step(
            point,
            target,
            proposal,
            n_particles=n_particles,
            # logp_x=logp_x,
            # logq_x=logq_x,
        )
        logp_x = log_ps[np.arange(point.shape[0]), indices]
        logq_x = log_qs[np.arange(point.shape[0]), indices]
        meta["sir_accept"].append((indices != 0).float().mean().item())
        point = point.detach().requires_grad_()
        meta["logp"] = logp_x
        # meta["mask"] = (
        #     F.one_hot(indices, num_classes=n_particles).to(bool).detach().cpu()
        # )
        if n_mala_steps > 0:
            points, meta = mala(
                point,
                target,
                proposal.prior,
                n_mala_steps,
                n_mala_steps - 1,
                project,
                step_size=step_size,
                target_acceptance=target_acceptance,
                verbose=False,
                meta=meta,
            )
            step_size = meta["step_size"][-1]
            point = points[-1]

        # meta.pop("logp")
        # meta.pop("grad")

        # if not keep_graph:
        point = point.detach().requires_grad_()
        if step_id >= burn_in:
            chains.append(point.cpu().clone())

        if proposal.optim.param_groups[0]["lr"] > 0:
            logw = log_ps - log_qs
            # KL(P | Q), MC with flow samples, Rao-Blackwellized
            kl_forw = (
                -(log_qs * torch.softmax(logw.detach(), dim=-1)).sum(axis=-1).mean()
            )
            # KL(Q | P), MC with current flow samples
            kl_back = -logw[:, 1:].mean()

            loss = forward_kl_weight * kl_forw + backward_kl_weight * kl_back

            # loss -= log_qs[:, 1:].mean() / kl_back.detach().item()
            proposal.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(proposal.parameters(), 1.0)
            proposal.optim.step()
            proposal.scheduler.step()

            meta["forward_kl"].append(kl_forw.item())
            meta["backward_kl"].append(kl_back.item())

            if verbose:
                pbar.set_description(
                    f"KL forw {kl_forw.item():.3f}, \
                     KL back {kl_back.item():.3f} "  # Hentr {entr.item():.3f}"
                )

    chains = torch.stack(chains, 0)

    return chains, meta



import copy
from typing import Callable, Tuple

import numpy as np
import torch
from torch import nn
from tqdm import trange

from .adaptive_sir_loss import get_loss
from .ebm_sampling import MALATransition, MHKernel, grad_energy
from .mcmc_base import AbstractMCMC, adapt_stepsize_dec, increment_steps


class AugmentedULA(AbstractMCMC):
    def __init__(self, flow=None, ula_steps=1, beta=1.0, **kwargs):
        super().__init__(**kwargs)
        self.flow = flow
        self.ula_steps = ula_steps
        self.beta = beta
        self.grad_step = kwargs.get("grad_step", 1e-2)
        self.noise_scale = kwargs.get(
            "noise_scale",
            (2 * self.grad_step) ** 0.5,
        )

    def __call__(
        self,
        start,
        target,
        proposal,
        n_steps=1,
        flow=None,
        ula_steps=None,
        beta=None,
        grad_step=None,
        noise_scale=None,
        **kwargs,
    ):
        flow = flow if flow is not None else self.flow
        ula_steps = ula_steps if ula_steps is not None else self.ula_steps
        beta = beta if beta is not None else self.beta
        grad_step = grad_step if grad_step is not None else self.grad_step
        noise_scale = noise_scale if noise_scale is not None else self.noise_scale

        zs = []
        batch_size = start.shape[0]
        device = start.device
        acc_rate = torch.zeros(batch_size)

        if flow is not None:
            z, _ = flow.inverse(start)
        else:
            z = start

        for step_id in range(n_steps):
            with torch.no_grad():
                x_new = proposal.sample((batch_size,))
                if flow is not None:
                    z_new, log_jac = flow.inverse(x_new)
                    x, minus_log_jac = flow.forward(z)
                    log_prop1 = proposal.log_prob(x) + minus_log_jac
                    log_prop2 = proposal.log_prob(x_new) - log_jac
                else:
                    pass
                log_tar1 = beta * target.log_prob(z)
                log_tar2 = beta * target.log_prob(z_new)
                mask = MHKernel.get_mask(
                    log_tar1,
                    log_tar2,
                    log_prop2,
                    log_prop1,
                )
                z[mask] = z_new[mask].detach().clone()
                z = z.data
                acc_rate += mask

            for _ in range(ula_steps):
                _, grad = grad_energy(z, target)
                z = (
                    z
                    - grad_step * beta * grad
                    + noise_scale * torch.randn(grad.shape, device=device)
                )
            zs.append(z.detach().clone())

        acc_rate /= n_steps

        return zs, acc_rate


class CorrelatedKernel:
    def __init__(self, corr_coef=0, bernoulli_prob_corr=0):
        self.corr_coef = corr_coef
        self.bern = torch.distributions.Bernoulli(bernoulli_prob_corr)

    def __call__(self, z, proposal, N, batch_size=1):
        correlation = self.corr_coef * self.bern.sample((batch_size,))  # (batch_size,)

        latent_var = correlation[:, None] * z + (
            1.0 - correlation[:, None] ** 2
        ) ** 0.5 * proposal.sample(
            (batch_size,),
        )  # (batch_size, z_dim)
        correlation_new = self.corr_coef * self.bern.sample(
            (batch_size, N)
        )  # (batch_size, N)
        z_new = correlation_new[..., None] * latent_var[:, None, :] + (
            1.0 - correlation_new[..., None] ** 2
        ) ** 0.5 * proposal.sample(
            (
                batch_size,
                N,
            ),
        )  # (batch_size, N, z_dim)
        return z_new


def compute_sir_log_weights(x, target, proposal, flow, beta=1.0):
    x_pushed, log_jac_inv = flow.inverse(x)
    log_weights = beta * target.log_prob(x_pushed) + log_jac_inv - proposal.log_prob(x)
    return log_weights, x_pushed


def adaptive_sir_correlated_dynamics(
    z,
    target,
    proposal,
    n_steps,
    N,
    corr_coef=0.0,
    bernoulli_prob_corr=0.0,
    flow=None,
    verbose=True,
):  # z assumed from proposal !
    z_sp = []  # vector of samples
    batch_size, z_dim = z.shape[0], z.shape[1]
    acceptance = torch.zeros(batch_size)

    range_gen = trange if verbose else range

    corr_ker = CorrelatedKernel(corr_coef, bernoulli_prob_corr)

    for _ in range_gen(n_steps):
        if flow is not None:
            z_pushed, _ = flow.inverse(z)
        else:
            z_pushed = z

        z_sp.append(z_pushed)

        X = corr_ker(z, proposal, N, batch_size)

        X[np.arange(batch_size), [0] * batch_size, :] = z
        X_view = X.view(-1, z_dim)

        if flow is not None:
            log_weight, z_pushed = compute_sir_log_weights(
                X_view,
                target,
                proposal,
                flow,
            )
        else:
            z_pushed = X_view
            log_weight = target.log_prob(z_pushed) - proposal.log_prob(z_pushed)

        log_weight = log_weight.view(batch_size, N)
        max_logs = torch.max(log_weight, dim=1)[0][:, None]
        log_weight = log_weight - max_logs
        weight = torch.exp(log_weight)
        sum_weight = torch.sum(weight, dim=1)
        weight = weight / sum_weight[:, None]

        weight[weight != weight] = 0.0
        weight[weight.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weight, 1).squeeze().tolist()
        mask = indices == 0
        acceptance += mask

        z = X[np.arange(batch_size), indices, :]

    if flow is not None:
        z_pushed, _ = flow.inverse(z)
    else:
        z_pushed = z

    z_sp.append(z_pushed)
    acceptance /= n_steps

    return z_sp, acceptance


class CISIR(AbstractMCMC):
    def __init__(
        self,
        N=2,
        corr_coef=0.0,
        bernoulli_prob_corr=0.0,
        flow=None,
        verbose=True,
        **kwargs,
    ):
        super().__init__()
        self.N = N
        self.corr_coef = corr_coef
        self.bernoulli_prob_corr = bernoulli_prob_corr
        self.flow = flow
        self.verbose = verbose
        self.n_steps = kwargs.get("n_steps", 1)  #

    # @increment_steps
    # @adapt_stepsize_dec
    def __call__(
        self,
        start: torch.Tensor,
        target,
        proposal,
        n_steps=None,
        N=None,
        corr_coef=None,
        bernoulli_prob_corr=None,
        flow=None,
        verbose=None,
    ):
        N = self.N if N is None else N
        corr_coef = self.corr_coef if corr_coef is None else corr_coef
        bernoulli_prob_corr = (
            self.bernoulli_prob_corr
            if bernoulli_prob_corr is None
            else bernoulli_prob_corr
        )
        flow = self.flow if flow is None else flow
        n_steps = self.n_steps if n_steps is None else n_steps
        verbose = self.verbose if verbose is None else verbose

        return adaptive_sir_correlated_dynamics(
            start,
            target,
            proposal,
            n_steps,
            N,
            corr_coef,
            bernoulli_prob_corr,
            flow,
            verbose,
        )


def ex2_mcmc_mala_new_proposal(
    z,
    target,
    proposal,
    n_steps,
    N,
    grad_step,
    noise_scale,
    beta=1.0,
    mala_steps=5,
    corr_coef=0.0,
    bernoulli_prob_corr=0.0,
    flow=None,
    adapt_stepsize=True,
    verbose=False,
    ind_chains=True,
):
    z_sp = []
    batch_size, z_dim = z.shape[0], z.shape[1]

    mala_transition = MALATransition(z_dim, z.device)
    mala_transition.adapt_grad_step = grad_step
    mala_transition.adapt_sigma = noise_scale

    corr_ker = CorrelatedKernel(corr_coef, bernoulli_prob_corr)

    if ind_chains:
        acceptance = torch.zeros(batch_size)
    else:
        acceptance = torch.zeros(batch_size * N)

    range_gen = trange if verbose else range

    if flow is not None:
        z_pushed, log_jac = flow.inverse(z)
    else:
        z_pushed = z

    for step_id in range_gen(n_steps):
        # z_sp.append(z_pushed.detach().cpu())

        if corr_coef == 0 and bernoulli_prob_corr == 0:
            X = proposal.sample([batch_size, N])
        else:
            X = corr_ker(z, proposal, N, batch_size)

        ind = [0] * batch_size
        X[np.arange(batch_size), ind, :] = z

        X_view = X.view(-1, z_dim)

        if flow is not None:
            log_weight, z_pushed = compute_sir_log_weights(
                X_view,
                target,
                proposal,
                flow,
                beta=beta,
            )
        else:
            z_pushed = X_view
            log_weight = beta * target.log_prob(X_view) - proposal.log_prob(X_view)

        log_weight = log_weight.view(batch_size, N)
        z_pushed = z_pushed.view(X.shape)

        max_logs = torch.max(log_weight, dim=1)[0][:, None]
        log_weight = log_weight - max_logs
        weight = torch.exp(log_weight)
        sum_weight = torch.sum(weight, dim=1)
        weight = weight / sum_weight[:, None]

        weight[weight != weight] = 0.0
        weight[weight.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weight, 1).squeeze().tolist()

        if ind_chains:
            z_pushed = z_pushed[np.arange(batch_size), indices, :]
        else:
            z_pushed = (
                z_pushed[np.arange(batch_size), indices, :][:, None, :]
                .repeat(1, N, 1)
                .reshape(batch_size * N, z_dim)
            )  # z_pushed.reshape(batch_size * N, z_dim)

        if step_id != n_steps - 1:
            if flow is not None:
                z, _ = flow.forward(z_pushed)
            else:
                z = z_pushed

        for _ in range(mala_steps):
            # compute values of energy and gradient of the energy function
            E, grad = grad_energy(z_pushed, target)
            z_pushed, _, _, mask = mala_transition(
                z_pushed,
                E,
                grad,
                target=target,
                adapt_stepsize=adapt_stepsize,
                beta=beta,
            )
            acceptance += mask.float() / mala_steps

        if not ind_chains:
            z = z_pushed.reshape(batch_size, N, z_dim)[np.arange(batch_size), 0, :]

        z_sp.append(z_pushed.detach().cpu())

    # z_sp.append(z_pushed.detach().cpu())
    acceptance /= n_steps

    return z_sp, acceptance, mala_transition.adapt_grad_step


def ex2_mcmc_mala(
    z,
    target,
    proposal,
    n_steps,
    N,
    grad_step,
    noise_scale,
    beta=1.0,
    mala_steps=5,
    corr_coef=0.0,
    bernoulli_prob_corr=0.0,
    device="cpu",
    flow=None,
    adapt_stepsize=True,
    verbose=False,
    ind_chains=True,
):
    z_sp = []
    batch_size, z_dim = z.shape[0], z.shape[1]

    mala_transition = MALATransition(z_dim, z.device)
    mala_transition.adapt_grad_step = grad_step
    mala_transition.adapt_sigma = noise_scale

    corr_ker = CorrelatedKernel(corr_coef, bernoulli_prob_corr)

    if ind_chains:
        acceptance = torch.zeros(batch_size)
    else:
        acceptance = torch.zeros(batch_size * N)

    range_gen = trange if verbose else range

    if flow is not None:
        z_pushed, _ = flow.inverse(z)
    else:
        z_pushed = z

    for step_id in range_gen(n_steps):
        # z_sp.append(z_pushed.detach().cpu())

        if corr_coef == 0 and bernoulli_prob_corr == 0:
            X = proposal.sample([batch_size, N])
        else:
            X = corr_ker(z, proposal, N, batch_size)

        ind = [0] * batch_size
        X[np.arange(batch_size), ind, :] = z

        X_view = X.view(-1, z_dim)

        if flow is not None:
            log_weight, z_pushed = compute_sir_log_weights(
                X_view,
                target,
                proposal,
                flow,
                beta=beta,
            )
        else:
            z_pushed = X_view
            log_weight = beta * target.log_prob(X_view) - proposal.log_prob(X_view)

        log_weight = log_weight.view(batch_size, N)
        z_pushed = z_pushed.view(X.shape)

        max_logs = torch.max(log_weight, dim=1)[0][:, None]
        log_weight = log_weight - max_logs
        weight = torch.exp(log_weight)
        sum_weight = torch.sum(weight, dim=1)
        weight = weight / sum_weight[:, None]

        weight[weight != weight] = 0.0
        weight[weight.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weight, 1).squeeze().tolist()

        if ind_chains:
            z_pushed = z_pushed[np.arange(batch_size), indices, :]
        else:
            z_pushed = (
                z_pushed[np.arange(batch_size), indices, :][:, None, :]
                .repeat(1, N, 1)
                .reshape(batch_size * N, z_dim)
            )  # z_pushed.reshape(batch_size * N, z_dim)

        for _ in range(mala_steps):
            E, grad = grad_energy(z_pushed, target)
            z_pushed, _, _, mask = mala_transition(
                z_pushed,
                E,
                grad,
                target=target,
                adapt_stepsize=adapt_stepsize,
                beta=beta,
            )
            acceptance += mask.cpu().float() / mala_steps

        if step_id != n_steps - 1:
            if flow is not None:
                z, _ = flow.forward(z_pushed)
            else:
                z = z_pushed

        if not ind_chains:
            z = z_pushed.reshape(batch_size, N, z_dim)[np.arange(batch_size), 0, :]

        z_sp.append(z_pushed.detach().cpu())
        # if step_id % 100 == 0:
        # gc.collect()
        # torch.cuda.empty_cache()

    # z_sp.append(z_pushed.detach().cpu())
    acceptance /= n_steps

    return z_sp, acceptance, mala_transition.adapt_grad_step


class Ex2MCMC(AbstractMCMC):
    def __init__(
        self,
        N=2,
        grad_step=1e-2,
        noise_scale=None,
        beta=1.0,
        mala_steps=5,
        corr_coef=0.0,
        bernoulli_prob_corr=0.0,
        device="cpu",
        flow=None,
        adapt_stepsize=False,
        verbose=True,
        ind_chains=True,
        **kwargs,
    ):
        super().__init__()
        self.N = N
        self.grad_step = grad_step
        self.noise_scale = (
            (2 * grad_step) ** 0.5 if noise_scale is None else noise_scale
        )
        self.mala_steps = mala_steps
        self.corr_coef = corr_coef
        self.bernoulli_prob_corr = bernoulli_prob_corr
        self.device = device
        self.flow = flow
        self.adapt_stepsize = adapt_stepsize
        self.verbose = verbose
        self.ind_chains = ind_chains
        self.n_steps = kwargs.get("n_steps", 1)  #
        self.beta = beta

    @increment_steps
    @adapt_stepsize_dec
    def __call__(self, start: torch.Tensor, target, proposal, *args, **kwargs):
        self_kwargs = copy.copy(self.__dict__)
        self_kwargs.update(kwargs)
        # gc.collect()
        # torch.cuda.empty_cache()
        n_steps = self_kwargs.pop("n_steps")
        if len(args) > 0:
            n_steps = args[0]

        self_kwargs.pop("_steps_done")

        return ex2_mcmc_mala(start, target, proposal, n_steps, **self_kwargs)


class FlowMCMC:
    def __init__(self, target, proposal, device, flow, mcmc_call: callable, **kwargs):
        self.flow = flow
        self.proposal = proposal
        self.target = target
        self.device = device
        self.batch_size = kwargs.get("batch_size", 64)
        self.mcmc_call = mcmc_call
        self.grad_clip = kwargs.get("grad_clip", 1.0)
        self.jump_tol = kwargs.get("jump_tol", 1e6)
        optimizer = kwargs.get("optimizer", "adam")
        loss = kwargs.get("loss", "mix_kl")
        self.flow.to(self.device)
        if isinstance(loss, (Callable, nn.Module)):
            self.loss = loss
        elif isinstance(loss, str):
            self.loss = get_loss(loss)(self.target, self.proposal, self.flow)
        else:
            ValueError

        lr = kwargs.get("lr", 1e-3)
        wd = kwargs.get("wd", 1e-4)
        if isinstance(optimizer, torch.optim.Optimizer):
            self.optimizer = optimizer
        elif isinstance(optimizer, str):
            if optimizer.lower() == "adam":
                self.optimizer = torch.optim.Adam(
                    flow.parameters(), lr=lr, weight_decay=wd
                )

        self.loss_hist = []

    def train_step(self, inp=None, alpha=0.5, do_step=True, inv=True):
        if do_step:
            self.optimizer.zero_grad()
        if inp is None:
            inp = self.proposal.sample((self.batch_size,))
        elif inv:
            inp, _ = self.flow.forward(inp)
        # print("before mcmc call")
        out = self.mcmc_call(inp, self.target, self.proposal, flow=self.flow)
        if isinstance(out, Tuple):
            acc_rate = out[1].mean()
            out = out[0]
        else:
            acc_rate = 1
        out = out[-1]
        out = out.to(self.device)
        nll = -self.target.log_prob(out).mean().item()

        if do_step:
            loss_est, loss = self.loss(out, acc_rate=acc_rate, alpha=alpha)

            if (
                len(self.loss_hist) > 0
                and loss.item() - self.loss_hist[-1] > self.jump_tol
            ) or torch.isnan(loss):
                print("KL wants to jump, terminating learning")
                return out, nll

            self.loss_hist = self.loss_hist[-500:] + [loss_est.item()]
            self.optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(
                self.flow.parameters(),
                self.grad_clip,
            )
            self.optimizer.step()

        return out, nll

    def train(self, n_steps=100, start_optim=10, init_points=None, alpha=None):
        samples = []
        inp = self.proposal.sample((self.batch_size,))

        neg_log_likelihood = []

        for step_id in trange(n_steps):
            # if alpha is not None:
            #    if isinstance(alpha, Callable):
            #        a = alpha(step_id)
            #    elif isinstance(alpha, float):
            #        a = alpha
            # else:
            a = min(0.5, 3 * step_id / n_steps)
            # print("before train step")
            out, nll = self.train_step(
                alpha=a,
                do_step=step_id >= start_optim,
                inp=init_points if step_id == 0 and init_points is not None else inp,
                inv=True,
            )
            inp = out.detach().requires_grad_()
            samples.append(inp.detach().cpu())

            neg_log_likelihood.append(nll)

        return samples, neg_log_likelihood

    def sample(self):
        pass



import numpy as np
import ot
import torch
import torch.nn as nn
import torch.nn.functional as F
from easydict import EasyDict as edict
from scipy.stats import chi2, entropy, gaussian_kde
from sklearn import mixture


@torch.no_grad()
def get_pis_estimate(
    X_gen,
    target_log_prob,
    n_pts=4000,
    sample_method="grid",
    density_method="gmm",
):
    target_pdf = (
        lambda x: target_log_prob(torch.FloatTensor(x)).exp().detach().cpu().numpy()
    )

    if density_method == "kde":
        G_ker = gaussian_kde(X_gen.detach().cpu().numpy().reshape(2, -1))

        def pi_g_f(x):
            return G_ker.pdf(x.reshape(2, -1))

    elif density_method == "gmm":
        gm_g = mixture.GaussianMixture(n_components=25)
        gm_g.fit(X_gen.detach().cpu().numpy())

        def pi_g_f(x):
            return np.exp(gm_g.score_samples(x))

    def optD(x):
        return target_pdf(x) / (target_pdf(x) + pi_g_f(x) + 1e-8)

    if sample_method == "grid":
        n_pts = int(n_pts**0.5)
        X = np.mgrid[-3 : 3 : 4.0 / n_pts, -3 : 3 : 4.0 / n_pts].reshape(2, -1).T
    elif sample_method == "mc":
        pass

    pi_d = target_pdf(X)
    pi_g = pi_g_f(X)

    opt_ds = optD(X)
    opt_ds[pi_d < 1e-9] = 0.0

    return pi_d, pi_g, opt_ds, X


class Evolution:
    def __init__(
        self,
        target_sample,
        **kwargs,
    ):  # locs, sigma, target_log_prob, target_sample, sigma=0.05, scaler=None):
        self.locs = kwargs.get("locs", None)
        self.sigma = kwargs.get("sigma", None)
        self.target_log_prob = kwargs.get("target_log_prob", None)
        self.target_sample = target_sample
        self.scaler = kwargs.get("scaler", None)
        self.q = kwargs.get("q", 0.999)

        self.mode_std = []
        self.high_quality_rate = []
        self.jsd = []
        self.emd = []
        self.n_found_modes = []

        self.kl_pis = []
        self.js_pis = []
        self.l2_div = []

    @staticmethod
    def make_assignment(
        X_gen,
        locs,
        sigma=0.05,
        q=0.95,
    ):
        n_modes, x_dim = locs.shape
        dists = torch.norm((X_gen[:, None, :] - locs[None, :, :]), p=2, dim=-1)
        chi2_quantile = chi2.ppf(q, x_dim)
        test_dist = chi2_quantile**0.5 * sigma
        assignment_chi2 = dists < test_dist
        assignment_2sigma = dists < 2 * sigma
        return assignment_chi2, assignment_2sigma

    @staticmethod
    def compute_mode_std(X_gen, assignment):
        """
        X_gen(torch.FloatTensor) - (n_pts, x_dim)

        """
        x_dim = X_gen.shape[-1]
        n_modes = assignment.shape[1]
        std = torch.FloatTensor([0.0]).to(X_gen.device)
        found_modes = 0
        for mode_id in range(n_modes):
            xs = X_gen[assignment[:, mode_id]]
            # print(xs.shape)
            if xs.shape[0] > 1:
                std_ = (
                    1 / (x_dim * (xs.shape[0] - 1)) * ((xs - xs.mean(0)) ** 2).sum()
                ) ** 0.5
                std += std_
                found_modes += 1
        if found_modes != 0:
            std /= found_modes
        return std, found_modes

    @staticmethod
    def compute_mode_mean(X_gen, assignment):
        """
        X_gen(torch.FloatTensor) - (n_pts, x_dim)
        """
        x_dim = X_gen.shape[-1]
        n_modes = assignment.shape[1]
        means = torch.zeros((n_modes, x_dim))
        found_modes_ind = []
        for mode_id in range(n_modes):
            xs = X_gen[assignment[:, mode_id]]
            # print(xs.shape)
            if xs.shape[0] > 1:
                means[mode_id] = xs.mean(0)
                found_modes_ind.append(mode_id)

        return means, found_modes_ind

    @staticmethod
    def compute_high_quality_rate(assignment):
        high_quality_rate = assignment.max(1)[0].sum() / float(
            assignment.shape[0],
        )
        return high_quality_rate

    @staticmethod
    def compute_jsd(assignment):
        device = assignment.device
        n_modes = assignment.shape[1]
        assign_ = torch.cat(
            [
                assignment,
                torch.zeros(assignment.shape[0]).to(device).unsqueeze(1),
            ],
            -1,
        )
        assign_[:, -1][assignment.sum(1) == 0] = 1
        sample_dist = assign_.sum(dim=0) / float(assign_.shape[0])
        sample_dist /= sample_dist.sum()
        uniform_dist = torch.FloatTensor(
            [1.0 / n_modes for _ in range(n_modes)] + [0],
        ).to(assignment.device)
        M = 0.5 * (uniform_dist + sample_dist)
        # JSD = .5 * (sample_dist * torch.log((sample_dist + 1e-7) / M)) + .5 * (uniform_dist * torch.log((uniform_dist + 1e-7) / M))

        # JSD[sample_dist == 0.] = 0.
        # JSD[uniform_dist == 0.] = 0.
        # JSD = JSD.sum()

        KL1 = sample_dist * (torch.log(sample_dist) - torch.log(M))
        KL1[sample_dist == 0.0] = 0.0
        KL2 = uniform_dist * (torch.log(uniform_dist) - torch.log(M))
        KL2[uniform_dist == 0.0] = 0.0
        JSD = 0.5 * (KL1 + KL2).sum()

        return JSD

    @staticmethod
    def compute_emd(target_sample, gen_sample):
        gen_sample = gen_sample[
            np.random.choice(
                np.arange(gen_sample.shape[0]),
                target_sample.shape[0],
                replace=target_sample.shape[0] > gen_sample.shape[0],
            )
        ]
        M = ot.dist(target_sample, gen_sample)
        emd = ot.emd2([], [], M)

        return emd

    def invoke(self, X_gen, D=None, compute_discrim_div=False):
        emd = Evolution.compute_emd(
            self.target_sample,
            X_gen.detach().cpu().numpy(),
        )
        self.emd.append(emd)

        if self.locs is not None and self.sigma is not None:
            assignment = Evolution.make_assignment(
                X_gen,
                self.locs,
                self.sigma,
                self.q,
            )
            # print(assignment.shape)
            mode_std, found_modes = Evolution.compute_mode_std(
                X_gen,
                assignment,
            )
            self.n_found_modes.append(found_modes)
            # print(mode_std)
            self.mode_std.append(mode_std.item())
            h_q_r = Evolution.compute_high_quality_rate(assignment)
            self.high_quality_rate.append(h_q_r.item())
            jsd = Evolution.compute_jsd(assignment)
            self.jsd.append(jsd.item())

        if self.target_log_prob is not None:
            pi_d, pi_g, opt_ds, X = get_pis_estimate(
                X_gen,
                self.target_log_prob,
                n_pts=4000,
                sample_method="grid",
                density_method="gmm",
            )
            kl = pi_g * (np.log(pi_g) - np.log(pi_d + 1e-10))
            kl[pi_g == 0.0] = 0.0
            kl = np.mean(kl).item()
            self.kl_pis.append(kl)

            m = 0.5 * (pi_g + pi_d)
            js = 0.5 * (
                pi_g * (np.log(pi_g) - np.log(m)) + pi_d * (np.log(pi_d) - np.log(m))
            )
            js[(pi_g == 0.0) + (pi_d == 0.0)] = 0.0
            js = np.mean(js).item()
            self.js_pis.append(js)

            if self.scaler is not None and compute_discrim_div and D is not None:
                ds = D(torch.FloatTensor(self.scaler.transform(X)))[:, 0]
                div = torch.abs(ds - torch.FloatTensor(opt_ds))
                # inf_dist = torch.max(div).item()
                l2_dist = ((div**2).mean()).item() ** 0.5
                self.l2_div.append(l2_dist)

    def as_dict(self):
        d = dict(
            mode_std=self.mode_std,
            hqr=self.high_quality_rate,
            jsd=self.jsd,
            emd=self.emd,
            kl_pis=self.kl_pis,
            js_pis=self.js_pis,
            l2_div=self.l2_div,
            n_found_modes=self.n_found_modes,
        )
        return d


def autocovariance(X, tau=0):
    # dT, dX = np.shape(X)
    dT = X.shape[0]
    s = 0.0
    dN = 1
    if tau > 0:
        x1 = X[:-tau, ...]
    else:
        x1 = X
    x2 = X[tau:, ...]
    s = np.sum(x1 * x2, axis=0) / dN

    return s / (dT - tau)


def acl_spectrum(X, n=150, scale=None):
    scale = np.array(scale) if scale is not None else np.sqrt(autocovariance(X, tau=0))
    return np.stack(
        [autocovariance(X / (scale[None, ...] + 1e-7), tau=t) for t in range(n - 1)],
        axis=0,
    )


def ESS(A):
    # ess = ESS(acl_spectrum((trunc_sample - trunc_sample.mean(0)[None, ...]))).mean()
    A = A * (A > 0.05)
    ess = 1.0 / (1.0 + 2 * np.sum(A[1:, ...], axis=0))
    return ess


class MetricsTracker:
    stor: dict

    def __init__(self, fields):
        self.fields = fields
        self.stor = edict()
        for field in self.fields:
            self.stor[field] = []



from pathlib import Path

import torch
from torch import nn

from ex2mcmc.utils.general_utils import DotConfig

from .base import MemoryModel, ModelRegistry


def init_weights(m):
    if isinstance(m, nn.Linear):
        torch.nn.init.xavier_uniform(m.weight)
        m.bias.data.fill_(0.01)


class CondDataParallel(torch.nn.DataParallel):
    label = None
    cond = False

    def forward(self, *inputs, **kwargs):
        if self.cond:
            return super().forward(*inputs, **kwargs, label=self.label)
        else:
            return super().forward(*inputs, **kwargs)


class GANWrapper:
    def __init__(
        self, config: DotConfig, device: torch.device, load_weights=True, eval=True
    ):
        self.config = config
        self.device = device

        self.gen = ModelRegistry.create(
            config.generator.name, **config.generator.params
        ).to(device)
        self.dis = ModelRegistry.create(
            config.discriminator.name, **config.discriminator.params
        ).to(device)

        # for n, p in self.gen.named_parameters():
        #     if 'weight' in n:
        #         torch.nn.init.orthogonal_(p, gain=0.8)

        if load_weights:
            self.load_weights()
        # else:
        #     self.gen.apply(init_weights)
        #     self.dis.apply(init_weights)

        if config.dp:
            self.gen = CondDataParallel(self.gen)
            self.dis = CondDataParallel(self.dis)
            self.dis.transform = self.dis.module.transform
            self.dis.output_layer = self.dis.module.output_layer
            self.gen.inverse_transform = self.gen.module.inverse_transform
            self.gen.z_dim = self.gen.module.z_dim
            self.gen.sample_label = self.gen.module.sample_label
            if hasattr(self.gen.module, "label"):
                self.gen.cond = self.gen.module.label
            if hasattr(self.dis.module, "label"):
                self.dis.cond = self.dis.module.label
            if hasattr(self.dis.module, "penult_layer"):
                self.dis.penult_layer = self.dis.module.penult_layer
        self.dp = config.dp

        if eval:
            self.gen = MemoryModel(self.gen)
            self.dis = MemoryModel(self.dis)

            dis_attrs = ["transform", "output_layer", "label", "penult_layer"]
            self.dis.__dict__.update(
                {attr: self.dis.module.__dict__.get(attr) for attr in dis_attrs}
            )
            gen_attrs = ["inverse_transform", "z_dim", "label"]
            self.gen.__dict__.update(
                {attr: self.gen.module.__dict__.get(attr) for attr in gen_attrs}
            )
            self.gen.sample_label = self.gen.module.sample_label
            if hasattr(self.gen.module, "label"):
                self.gen.cond = self.gen.module.label
            if hasattr(self.dis.module, "label"):
                self.dis.cond = self.dis.module.label

            self.dis.transform = self.dis.module.transform
        print(f"Transform: {self.dis.transform}")

        if eval:
            self.eval()
        self.define_prior()
        self.label = None

    def load_weights(self):
        from ex2mcmc.utils.general_utils import PROJECT_PATH

        gen_path = Path(PROJECT_PATH, self.config.generator.ckpt_path)
        # if not gen_path.exists():
        #     subprocess.run(["dvc pull", gen_path.parent])

        state_dict = torch.load(gen_path, map_location=self.device)
        try:
            self.gen.load_state_dict(state_dict, strict=True)
        except RuntimeError:
            state_dict = {k.replace("module.", ""): v for k, v in state_dict.items()}
            self.gen.load_state_dict(state_dict, strict=True)

        dis_path = Path(PROJECT_PATH, self.config.discriminator.ckpt_path)
        # if not dis_path.exists():
        #     subprocess.run(["dvc pull", dis_path.parent])
        state_dict = torch.load(
            dis_path,
            map_location=self.device,
        )
        try:
            self.dis.load_state_dict(state_dict, strict=True)
        except RuntimeError:
            state_dict = {k.replace("module.", ""): v for k, v in state_dict.items()}
            self.dis.load_state_dict(state_dict, strict=True)

    def eval(self):
        for param in self.gen.parameters():
            param.requires_grad = False
        for param in self.dis.parameters():
            param.requires_grad = False
        self.gen.eval()
        self.dis.eval()

    def get_latent_code_dim(self):
        return self.gen.z_dim

    def define_prior(self):
        if self.config.prior == "normal":
            prior = torch.distributions.multivariate_normal.MultivariateNormal(
                torch.zeros(self.gen.z_dim).to(self.device),
                torch.eye(self.gen.z_dim).to(self.device),
            )
            prior.project = lambda z: z
        elif self.config.prior == "uniform":
            prior = torch.distributions.uniform.Uniform(
                -torch.ones(self.gen.z_dim).to(self.device),
                torch.ones(self.gen.z_dim).to(self.device),
            )
            prior.project = lambda z: torch.clip(z, -1 + 1e-9, 1 - 1e-9)
            prior.log_prob = lambda z: torch.zeros(z.shape[0], device=z.device)
        else:
            raise KeyError
        self.gen.prior = prior

    @property
    def transform(self):
        return self.dis.transform

    @property
    def inverse_transform(self):
        return self.gen.inverse_transform

    @property
    def prior(self):
        return self.gen.prior

    def set_label(self, label):
        if self.config.dp:
            self.gen.label = label if self.gen.cond else None
            self.dis.label = label if self.dis.cond else None
            self.label = label if self.dis.cond else None
        else:
            self.gen.label = label
            self.dis.label = label
            self.label = label



"""
Code is partially borrowed from repository https://github.com/sbarratt/inception-score-pytorch/blob/master/inception_score.py # noqa: E501
"""

import argparse
import logging
from pathlib import Path
from typing import Dict, Iterable, Optional, Tuple, Union

import numpy as np
import torch
import torch.utils.data
import torchvision
import torchvision.datasets as dset
import torchvision.transforms as transforms
import yaml
from torch import nn
from torch.nn import functional as F
from torchvision.models.inception import inception_v3
from yaml import Loader

from ex2mcmc.utils.callbacks import Callback, CallbackRegistry


N_INCEPTION_CLASSES = 1000
MEAN_TRASFORM = [0.485, 0.456, 0.406]
STD_TRANSFORM = [0.229, 0.224, 0.225]
N_GEN_IMAGES = 100005000


@CallbackRegistry.register()
class InceptionScoreCallback(Callback):
    def __init__(
        self,
        invoke_every: int = 1,
        device: Union[str, int, torch.device] = "cuda",
        update_input: bool = True,
        dp: bool = False,
        batch_size: int = 128,
    ) -> None:
        self.device = device
        self.model = torchvision.models.inception.inception_v3(
            pretrained=True, transform_input=False
        ).to(device)
        if dp:
            self.model = nn.DataParallel(self.model)

        self.model.eval()
        self.transform = transforms.Normalize(mean=MEAN_TRASFORM, std=STD_TRANSFORM)
        self.update_input = update_input
        self.invoke_every = invoke_every
        self.batch_size = batch_size

    @torch.no_grad()
    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        score = None
        step = info.get("step", self.cnt)
        if step % self.invoke_every == 0:
            imgs = torch.from_numpy(info["imgs"]).to(self.device)
            imgs = self.transform(imgs)
            pis = []
            for batch in torch.split(imgs, self.batch_size):
                pis_ = batch_inception(batch, self.model, resize=True).detach().cpu()
                pis.append(pis_)
            pis = torch.cat(pis, 0)
            score = (
                (pis * (torch.log(pis) - torch.log(pis.mean(0)[None, :])))
                .sum(1)
                .mean(0)
            )
            score = torch.exp(score)

            if self.update_input:
                info["inception_score"] = score
            logger = logging.getLogger()
            logger.info(f"\nIS: {score}")
        self.cnt += 1
        return score


def batch_inception(
    imgs: torch.Tensor,
    inception_model: nn.Module,
    resize: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    device = imgs.device
    up = nn.Upsample(size=(299, 299), mode="bilinear").to(device)

    def get_pred(x):
        if resize:
            x = up(x)
        x = inception_model(x)
        return F.softmax(x, -1)

    preds = get_pred(imgs)
    return preds


def get_inception_score(
    imgs: Iterable,
    inception_model: Optional[nn.Module] = None,
    gen: Optional[nn.Module] = None,
    generate_from_latents: bool = False,
    cuda: bool = True,
    batch_size: int = 32,
    resize: bool = False,
    splits: int = 1,
    device: Union[torch.device, int] = 0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes the inception score of the generated images imgs
    imgs -- Torch dataset of (3xHxW) numpy images normalized in the range [-1, 1]
    cuda -- whether or not to run on GPU
    batch_size -- batch size for feeding into Inception v3
    splits -- number of splits
    """
    N = len(imgs)

    assert batch_size > 0
    assert N >= batch_size

    # Set up dataloader
    if not isinstance(imgs, torch.utils.data.Dataset):
        imgs = torch.utils.data.TensorDataset(imgs)

    dataloader = torch.utils.data.DataLoader(imgs, batch_size=batch_size)

    # Load inception model
    if inception_model is None:
        inception_model = inception_v3(
            pretrained=True,
            transform_input=False,  # False
        ).to(device)
        inception_model.eval()

    up = nn.Upsample(size=(299, 299), mode="bilinear").to(device)

    def get_pred(x):
        if resize:
            x = up(x)
        x = inception_model(x)
        return F.softmax(x, -1).data.cpu()

    # Get predictions
    preds = torch.zeros((N, N_INCEPTION_CLASSES))

    for i, batch in enumerate(dataloader):
        if isinstance(batch, list):
            batch = batch[0]
        batch = batch.to(device)
        if generate_from_latents:
            batch = gen(batch)
        batch_size_i = batch.size()[0]

        preds[i * batch_size : i * batch_size + batch_size_i] = get_pred(batch)

    # Now compute the mean kl-div
    split_scores = []

    for k in range(splits):
        part = preds[k * (N // splits) : (k + 1) * (N // splits), :]
        py = torch.mean(part, 0)
        split_scores.append(
            torch.exp(
                (part * (torch.log(part) - torch.log(py[None, :])))
                .sum(1)
                .mean(0)
                # torch.mean(torch.kl_div(part, torch.log(py[None, :])).sum(1))
            )
        )

    return (
        torch.mean(torch.stack(split_scores, 0)),
        torch.std(torch.stack(split_scores, 0)),
        preds,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--gan_config", type=str)
    parser.add_argument("--device", type=int, default=0)
    args = parser.parse_args()

    device = torch.device(args.device)

    if args.gan_config:
        from ex2mcmc.models.utils import GANWrapper
        from ex2mcmc.utils.general_utils import DotConfig

        gan_config = DotConfig(
            yaml.load(Path(args.gan_config).open("r"), Loader)
        ).gan_config
        # gen, _ = load_gan(gan_config, device)
        gan = GANWrapper(gan_config, device=device, load_weights=True)
        gen = gan.gen

        n_imgs = 1000  # 1000
        imgs = gen(torch.randn(n_imgs, gen.z_dim).to(device))
        imgs = gen.inverse_transform(imgs)

        transform = transforms.Normalize(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )
        imgs = transform(imgs)

        print("Calculating Inception Score...")
        print(
            get_inception_score(
                imgs,
                cuda=True,
                batch_size=32,
                resize=True,
                splits=1,
            )[:-1]
        )
    else:

        class IgnoreLabelDataset(torch.utils.data.Dataset):
            def __init__(self, orig):
                self.orig = orig

            def __getitem__(self, index):
                return self.orig[index][0]

            def __len__(self):
                return len(self.orig)

        cifar = dset.CIFAR10(
            root="data/",
            download=True,
            transform=transforms.Compose(
                [
                    transforms.Scale(32),
                    transforms.ToTensor(),
                    # transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                ]
            ),
        )

        IgnoreLabelDataset(cifar)

        print("Calculating Inception Score...")
        print(
            get_inception_score(
                IgnoreLabelDataset(cifar),
                cuda=True,
                batch_size=32,
                resize=True,
                splits=10,
            )[:-1]
        )



from dataclasses import dataclass

import jax
import numpy as np
from jax import numpy as jnp
from scipy.stats import gaussian_kde
from tqdm.auto import trange


class MeanTracker:
    def __init__(self):
        self.values = []

    def update(self, value: float) -> None:
        self.values.append(value)

    def __len__(self):
        return len(self.values)

    def mean(self) -> float:
        return jnp.mean(jnp.array(self.values))

    def std(self) -> float:
        return jnp.std(jnp.array(self.values), ddof=1)

    def std_of_mean(self) -> float:
        return jnp.std(jnp.array(self.values)) / jnp.sqrt(len(self))

    def last(self) -> float:
        return self.values[-1]


@dataclass
class Projector:
    x0: jnp.ndarray
    v: jnp.ndarray

    def project(self, xs: jnp.ndarray) -> jnp.ndarray:
        return (xs - self.x0[None]) @ self.v


def create_random_projection(key: jnp.ndarray, xs: jnp.ndarray) -> Projector:
    x0 = jnp.mean(xs, 0)
    v = jax.random.normal(key, [len(x0)])
    v = v / jnp.linalg.norm(v)

    return Projector(x0, v)


def create_random_2d_projection(
    key: jnp.ndarray,
    xs: jnp.ndarray,
) -> Projector:
    x0 = jnp.mean(xs, 0)
    v = jax.random.normal(key, [len(x0), len(x0)])
    v = v / jnp.linalg.norm(v)

    print(v.shape)

    return Projector(x0, v)


def average_total_variation(
    key: jnp.ndarray,
    true: jnp.ndarray,
    other: jnp.ndarray,
    n_samples: int,
    n_steps: int,
) -> MeanTracker:
    tracker = MeanTracker()
    keys = jax.random.split(key, n_steps)

    for b in range(other.shape[1]):
        for i in trange(n_steps, leave=False):
            tracker.update(total_variation(keys[i], true, other[:, b], n_samples))
    return tracker


def total_variation(
    key: jnp.ndarray,
    xs_true: jnp.ndarray,
    xs_pred: jnp.ndarray,
    n_samples: int,
):
    proj = create_random_projection(key, xs_true)
    return total_variation_1d(
        proj.project(xs_true),
        proj.project(xs_pred),
        n_samples,
    )


def total_variation_1d(xs_true, xs_pred, n_samples):
    true_density = gaussian_kde(xs_true)
    pred_density = gaussian_kde(xs_pred)

    x_min = min(xs_true.min(), xs_pred.min())
    x_max = max(xs_true.max(), xs_pred.max())

    points = np.linspace(x_min, x_max, n_samples)

    return (
        0.5
        * np.abs(true_density(points) - pred_density(points)).mean()
        * (x_max - x_min)
    )


# def average_emd(
#     key: jnp.ndarray, true: jnp.ndarray, other: jnp.ndarray, n_samples: int, n_steps: int
# ) -> MeanTracker:
#     tracker = MeanTracker()
#     keys = jax.random.split(key, n_steps)
#     for i in trange(n_steps, leave=False):
#         tracker.update(emd_2d(keys[i], true, other, n_samples))
#     return tracker


# def emd_2d(key: jnp.ndarray, xs_true: jnp.ndarray, xs_pred: jnp.ndarray, n_samples: int):
#     proj = create_random_2d_projection(key, xs_true)
#     return earth_movers_distance_2d(proj.project(xs_true), proj.project(xs_pred), n_samples)


# def earth_movers_distance_2d(xs_true, xs_pred, n_samples):
#     print(xs_true.shape, xs_pred.shape)
#     M = np.linalg.norm(xs_true[None, :, :] - xs_pred[:, None, :], axis=-1, ord=2)**2
#     emd = ot.lp.emd2([], [], M)
#     return emd



import numpy as np
import torch
import tqdm
from torch.distributions import Categorical
from torch.distributions import MultivariateNormal as MNormal


# MALA
def grad_log_and_output_target_dens(log_target_dens, z):
    """
    returns the gradient of log-density
    """
    inp = z.clone().detach()
    inp.requires_grad_(True)

    output = log_target_dens(inp, detach=False)
    output.sum().backward()

    grad = inp.grad.data.detach()

    return grad, output.data.detach()


def mala_step(log_target_dens, x0, gamma, mala_iters, stats=None):
    """
    function to perform mala_iters times MALA step
    """
    N_samples, lat_size = x0.shape
    device = x0.device

    # generate proposals
    x_cur = x0
    for i in range(mala_iters):
        x_cur_grad, x_cur_log_target = grad_log_and_output_target_dens(
            log_target_dens, x_cur
        )
        x_cur_proposal = MNormal(
            x_cur + gamma * x_cur_grad,
            2
            * gamma
            * torch.eye(lat_size)[None, :, :].repeat(N_samples, 1, 1).to(device),
        )

        x_next = x_cur_proposal.sample()

        x_next_grad, x_next_log_target = grad_log_and_output_target_dens(
            log_target_dens, x_next
        )
        x_next_proposal = MNormal(
            x_next + gamma * x_next_grad,
            2
            * gamma
            * torch.eye(lat_size)[None, :, :].repeat(N_samples, 1, 1).to(device),
        )

        # compute accept-reject
        log_prob_accept = (
            x_next_log_target
            + x_next_proposal.log_prob(x_cur)
            - x_cur_log_target
            - x_cur_proposal.log_prob(x_next)
        )
        log_prob_accept = torch.clamp(log_prob_accept, max=0.0)

        # generate uniform distribution
        trhd = torch.log(torch.rand(N_samples).to(device))
        # accept-reject indicator
        indic_acc = (log_prob_accept > trhd).float()

        indic_acc = indic_acc[:, None]

        if stats is not None:
            stats["n_accepts"] += indic_acc.sum().item()

        x_cur = indic_acc * x_next + (1 - indic_acc) * x_cur
    return x_cur.detach()


def mala(log_target_dens, x0, N_steps, gamma, mala_iters, stats=None, seed=42):
    torch.manual_seed(seed)
    np.random.seed(seed)
    ### sample i-sir

    samples_traj = [x0]
    x_cur = x0

    for _ in tqdm.tqdm(range(N_steps)):
        x_cur = mala_step(log_target_dens, x_cur, gamma, mala_iters, stats=stats)
        samples_traj.append(x_cur)

    samples_traj = torch.stack(samples_traj).transpose(0, 1)

    return samples_traj


# I-SIR
def i_sir_step(log_target_dens, x_cur, N_part, isir_proposal, return_all_stats=False):
    """
    function to sample with N-particles version of i-SIR
    args:
        N_part - number of particles, integer;
        x0 - current i-sir particle;
    return:
        x_next - selected i-sir particle
    """
    N_samples, lat_size = x_cur.shape

    # generate proposals
    proposals = isir_proposal.sample(
        (
            N_samples,
            N_part - 1,
        )
    )

    # put current particles
    proposals = torch.cat((x_cur[:, None, :], proposals), dim=1)

    # compute importance weights
    log_target_dens_proposals = log_target_dens(
        proposals.reshape(-1, lat_size)
    ).reshape(N_samples, N_part)

    logw = log_target_dens_proposals - isir_proposal.log_prob(proposals)

    # sample selected particle indexes
    idxs = Categorical(logits=logw).sample()

    cur_x = proposals[torch.arange(N_samples), idxs]

    cur_x = cur_x.detach()
    proposals = proposals.detach()
    log_target_dens_proposals = log_target_dens_proposals.detach()

    if return_all_stats:
        return cur_x, proposals, log_target_dens_proposals
    else:
        return cur_x


def mh_step(log_target_dens, x_cur, mh_proposal):
    """
    function to sample with N-particles version of i-SIR
    args:
        N_part - number of particles, integer;
        x0 - current i-sir particle;
    return:
        x_next - selected i-sir particle
    """
    N_samples, lat_size = x_cur.shape
    device = x_cur.device
    # generate proposals
    proposals = mh_proposal.sample((N_samples,))
    # put current particles
    log_target_cur = log_target_dens(x_cur)
    log_target_proposal = log_target_dens(proposals)
    # compute proposal density
    log_prop_cur = mh_proposal.log_prob(x_cur)
    log_prop_prop = mh_proposal.log_prob(proposals)
    # compute acceptance ratio
    log_prob_accept = (
        log_target_proposal + log_prop_cur - log_prop_prop - log_target_cur
    )
    log_prob_accept = torch.clamp(log_prob_accept, max=0.0)
    # generate uniform distribution
    trhd = torch.log(torch.rand(N_samples).to(device))
    # accept-reject indicator
    indic_acc = (log_prob_accept > trhd).float()
    indic_acc = indic_acc[:, None]
    cur_x = indic_acc * proposals + (1 - indic_acc) * x_cur
    return cur_x.detach()


def i_sir(log_target_dens, x0, N_steps, N_part, isir_proposal, seed=42):
    torch.manual_seed(seed)
    np.random.seed(seed)
    ### sample i-sir

    samples_traj = [x0]
    x_cur = x0

    for _ in tqdm.tqdm(range(N_steps)):
        x_cur = i_sir_step(log_target_dens, x_cur, N_part, isir_proposal)
        samples_traj.append(x_cur)

    samples_traj = torch.stack(samples_traj).transpose(0, 1)

    return samples_traj


# Ex2MCMC
def ex2_mcmc(
    log_target_dens,
    x0,
    N_steps,
    N_part,
    isir_proposal,
    gamma,
    mala_iters,
    stats=None,
    seed=42,
):
    torch.manual_seed(seed)
    np.random.seed(seed)
    ### sample i-sir

    samples_traj = [x0]
    x_cur = x0

    for _ in tqdm.tqdm(range(N_steps)):
        x_cur = i_sir_step(log_target_dens, x_cur, N_part, isir_proposal)
        x_cur = mala_step(log_target_dens, x_cur, gamma, mala_iters, stats=stats)
        samples_traj.append(x_cur)

    samples_traj = torch.stack(samples_traj).transpose(0, 1)

    return samples_traj


def flex2_mcmc(
    log_target_dens,
    x0,
    N_steps,
    N_part,
    isir_proposal,
    gamma,
    mala_iters,
    alpha=1.0,
    stats=None,
    seed=42,
):
    """
    Function to perform Flex2MCMC by optimizing alpha * KL_FW + (1-alpha)*KL_BACK
    """
    torch.manual_seed(seed)
    np.random.seed(seed)
    ### sample i-sir
    samples_traj = [x0]
    x_cur = x0
    proposal_opt = torch.optim.Adam(isir_proposal.parameters(), lr=1e-3)

    pbar = tqdm.tqdm(range(N_steps))

    for _ in pbar:
        x_cur, proposals, log_target_dens_proposals = i_sir_step(
            log_target_dens, x_cur, N_part, isir_proposal, return_all_stats=True
        )
        # train proposal
        proposal_opt.zero_grad()
        population_log_proposal_prob = isir_proposal.log_prob(proposals)
        logw = (log_target_dens_proposals - population_log_proposal_prob).detach()

        kl_forw = -(
            (population_log_proposal_prob * torch.softmax(logw, dim=-1)).sum(axis=-1)
        ).mean()

        # backward KL
        cur_samples_x = proposals[:, 1:, :].reshape(-1, proposals.shape[-1])
        cur_samples_z, log_det_J = isir_proposal.f(cur_samples_x)

        cur_samples_x_diff = isir_proposal.g(cur_samples_z.detach())[0]
        log_target_dens_proposals_diff = log_target_dens(
            cur_samples_x_diff, detach=False
        )

        kl_back = -(log_target_dens_proposals_diff - log_det_J).mean()

        # opt step
        loss = alpha * kl_forw + (1 - alpha) * kl_back
        loss.backward()
        proposal_opt.step()

        # perform MALA update
        x_cur = mala_step(log_target_dens, x_cur, gamma, mala_iters, stats=stats)
        samples_traj.append(x_cur)

        pbar.set_description(f"KL forw {kl_forw.item()}, KL back {kl_back.item()}")

    samples_traj = torch.stack(samples_traj).transpose(0, 1)
    return samples_traj



import torch
from pyro.infer import HMC as pyro_hmc
from pyro.infer import MCMC
from pyro.infer import NUTS as pyro_nuts


def NUTS(
    start,
    target,
    n_samples: int,
    burn_in: int,
    *,
    step_size: float,
    verbose: bool = False,
) -> torch.FloatTensor:
    """
    No-U-Turn Sampler

    Args:
        start - strating points of shape [n_chains x dim]
        target - target distribution instance with method "log_prob"
        n_samples - number of last samples from each chain to return
        burn_in - number of first samples from each chain to throw away
        step_size - step size for drift term
        verbose - whether to show iterations' bar

    Returns:
        tensor of chains with shape [n_samples, n_chains, dim], acceptance rates
        for each iteration
    """
    x = start.clone()
    x.requires_grad_(False)

    def energy(z):
        z = z["points"]
        return -target.log_prob(z).sum()

    kernel = pyro_nuts(potential_fn=energy, step_size=step_size, full_mass=False)

    init_params = {"points": x}
    mcmc_true = MCMC(
        kernel=kernel,
        num_samples=n_samples,
        initial_params=init_params,
        warmup_steps=burn_in,
    )
    mcmc_true.run()

    q_true = mcmc_true.get_samples(group_by_chain=True)["points"]
    samples_true = q_true.view(-1, *start.shape)

    return samples_true


def HMC(
    start,
    target,
    n_samples: int,
    burn_in: int,
    *,
    step_size: float,
    num_leapfrog_steps: float = 1,
    verbose: bool = False,
) -> torch.FloatTensor:
    """
    Hamiltonian Monte Carlo

    Args:
        start - strating points of shape [n_chains x dim]
        target - target distribution instance with method "log_prob"
        n_samples - number of last samples from each chain to return
        burn_in - number of first samples from each chain to throw away
        step_size - step size for drift term
        verbose - whether to show iterations' bar

    Returns:
        tensor of chains with shape [n_samples, n_chains, dim], acceptance rates
        for each iteration
    """
    x = start.clone()
    x.requires_grad_(False)

    def energy(z):
        z = z["points"]
        return -target.log_prob(z).sum()

    kernel = pyro_hmc(
        potential_fn=energy,
        step_size=step_size,
        num_steps=num_leapfrog_steps,
        full_mass=False,
    )

    init_params = {"points": x}
    mcmc_true = MCMC(
        kernel=kernel,
        num_samples=n_samples,
        initial_params=init_params,
        warmup_steps=burn_in,
    )
    mcmc_true.run()

    q_true = mcmc_true.get_samples(group_by_chain=True)["points"]
    samples_true = q_true.view(-1, *start.shape)

    return samples_true



from abc import ABC, abstractmethod

import numpy as np
import numpy.random as rng
import torch
import torch.distributions as td
import torch.nn.functional as F
from easydict import EasyDict as edict
from matplotlib import pyplot as plt
from torch import nn

from .linear_regression import RegressionDataset
from .logistic_regression import ClassificationDataset


torchType = torch.float32


class Distribution(ABC):
    """
    Base class for a custom target distribution
    """

    def __init__(self, **kwargs):
        super().__init__()
        self.device = kwargs.get("device", "cpu")
        self.torchType = torchType
        self.xlim, self.ylim = [-1, 1], [-1, 1]
        self.scale_2d_log_prob = 1
        # self.device_zero = torch.tensor(0., dtype=self.torchType, device=self.device)
        # self.device_one = torch.tensor(1., dtype=self.torchType, device=self.device)

    def prob(self, x):
        """
        The method returns target density, estimated at point x
        Input:
        x - datapoint
        Output:
        density - p(x)
        """
        # You should define the class for your custom distribution
        return self.log_prob(x).exp()

    @abstractmethod
    def log_prob(self, x):
        """
        The method returns target logdensity, estimated at point x
        Input:
        x - datapoint
        Output:
        log_density - log p(x)
        """
        # You should define the class for your custom distribution
        raise NotImplementedError

    def energy(self, x):
        """
        The method returns target logdensity, estimated at point x
        Input:
        x - datapoint
        Output:
        energy = -log p(x)
        """
        # You should define the class for your custom distribution
        return -self.log_prob(x)

    def sample(self, n):
        """
        The method returns samples from the distribution
        Input:
        n - amount of samples
        Output:
        samples - samples from the distribution
        """
        # You should define the class for your custom distribution
        raise NotImplementedError

    def __call__(self, x):
        return self.log_prob(x)

    def log_prob_2d_slice(self, z):
        raise NotImplementedError

    def plot_2d(self, fig=None, ax=None):
        if fig is None and ax is None:
            fig, ax = plt.subplots()

        x = np.linspace(*self.xlim, 100)
        y = np.linspace(*self.ylim, 100)
        xx, yy = np.meshgrid(x, y)
        z = torch.FloatTensor(np.stack([xx, yy], -1))
        vals = (self.log_prob_2d_slice(z) / self.scale_2d_log_prob).exp()

        if ax is not None:
            ax.imshow(
                vals.flip(0),
                extent=[*self.xlim, *self.ylim],
                cmap="Greens",
                alpha=0.5,
                aspect="auto",
            )
        else:
            plt.imshow(
                vals.flip(0),
                extent=[*self.xlim, *self.ylim],
                cmap="Greens",
                alpha=0.5,
                aspect="auto",
            )

        return fig, self.xlim, self.ylim


class GaussianMixture(Distribution):
    """
    Mixture of n gaussians (multivariate)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.locs = kwargs.get(
            "locs",
            torch.FloatTensor([[-1, 0], [1, 0]]),
        )  # list of locations for each of these gaussians
        self.num = kwargs.get("num_gauss", len(self.locs))
        self.pis = kwargs.get(
            "p_gaussians",
            torch.FloatTensor([1.0 / self.num] * self.num),
        )
        self.dim = kwargs.get("dim", self.locs.shape[1])
        self.sigma = kwargs.get("sigma", 0.2)
        self.covs = kwargs.get(
            "covs",
            [self.sigma**2 * torch.eye(self.dim)] * self.num,
        )  # list of covariance matrices for each of these gaussians

        self.peak = [None] * self.num
        for i in range(self.num):
            self.peak[i] = torch.distributions.MultivariateNormal(
                loc=self.locs[i],
                covariance_matrix=self.covs[i],
            )

    def get_density(self, x):
        """
        The method returns target density
        Input:
        x - datapoint
        z - latent vaiable
        Output:
        density - p(x)
        """
        density = self.log_prob(x).exp()
        return density

    def log_prob(self, z):
        """
        The method returns target logdensity
        Input:
        x - datapoint
        z - latent variable
        Output:
        log_density - log p(x)
        """
        log_p = torch.tensor([], device=self.device)
        # pdb.set_trace()
        for i in range(self.num):
            log_paux = (
                torch.log(self.pis[i]) + self.peak[i].log_prob(z.to(self.device))
            ).view(-1, 1)
            log_p = torch.cat([log_p, log_paux], dim=-1)
        log_density = torch.logsumexp(log_p, dim=1)
        return log_density.view(z.shape[:-1])

    def energy(self, z, x=None):
        return -self.log_prob(z, x)

    def plot_2d(self):
        if self.dim != 2:
            raise NotImplementedError("can't plot for gaussians not in 2d")

        fig, ax = plt.subplots()

        dim1 = 0
        dim2 = 1

        xlim = [
            self.locs[:, dim1].min().item() - 1,
            self.locs[:, dim1].max().item() + 1,
        ]
        ylim = [
            self.locs[:, dim2].min().item() - 1,
            self.locs[:, dim2].max().item() + 1,
        ]

        x = np.linspace(*xlim, 100)
        y = np.linspace(*ylim, 100)
        xx, yy = np.meshgrid(x, y)
        z = torch.FloatTensor(np.stack([xx, yy], -1))
        vals = self.log_prob(z).exp()

        plt.contourf(
            vals,
            extent=[*xlim, *ylim],
            cmap="Greens",
            alpha=0.5,
            aspect="auto",
        )

        return fig, xlim, ylim


class IndependentNormal(Distribution):
    def __init__(self, dim, **kwargs):
        super().__init__(**kwargs)
        self.device = kwargs.get("device", "cpu")
        self.dim = dim
        self.loc = kwargs.get("loc", torch.zeros(self.dim))
        self.scale = kwargs.get("scale", 1.0)
        self.distribution = torch.distributions.Normal(
            loc=self.loc,
            scale=self.scale,
        )

    def log_prob(self, z, x=None):
        log_density = (self.distribution.log_prob(z.to(self.device))).sum(
            dim=-1,
        )
        return log_density

    def sample(self, n):
        return self.distribution.sample(n)

    def energy(self, z, x=None):
        return -self.log_prob(z, x)


def init_independent_normal(scale, n_dim, device, loc=0.0):
    loc = loc * torch.ones(n_dim).to(device)
    scale = scale * torch.ones(n_dim).to(device)
    target_args = edict()
    target_args.device = device
    target_args.loc = loc
    target_args.scale = scale
    target = IndependentNormal(n_dim, **target_args)
    return target


def init_independent_normal_scale(scales, locs, device):
    target_args = edict()
    target_args.device = device
    target_args.loc = locs.to(device)
    target_args.scale = scales.to(device)
    target = IndependentNormal(**target_args)
    return target


class Cauchy(Distribution):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.loc = kwargs.get("loc")
        self.scale = kwargs.get("scale")
        self.dim = kwargs.get("dim")
        self.distr = torch.distributions.Cauchy(self.loc, self.scale)

    def log_prob(self, z, x=None):
        log_target = self.distr.log_prob(z)  # .sum(-1)

        return log_target

    def sample(self, n=(1,)):
        return self.distr.sample((n[0], self.dim))


class CauchyMixture(Distribution):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.device = kwargs.get("device", "cpu")
        self.dim = kwargs.get("dim", 48)
        self.mu = kwargs.get("mu", 2 * torch.ones(self.dim))
        self.cov = kwargs.get("cov", 0.7 * torch.ones(self.dim))
        self.cauchy = Cauchy(loc=self.mu, scale=self.cov)
        self.cauchy_minus = Cauchy(loc=-self.mu, scale=self.cov)
        self.xlim = [-5, 5]
        self.ylim = [-5, 5]

    def get_density(self, x):
        return self.log_prob(x).exp()

    def log_prob_2d_slice(self, z, dim1=0, dim2=1):
        mu = self.mu[[dim1, dim2]]
        cov = self.cov[[dim1, dim2]]

        cauchy = Cauchy(loc=mu, scale=cov)
        cauchy_minus = Cauchy(loc=-mu, scale=cov)

        catted = torch.cat(
            [
                cauchy.log_prob(z)[None, ...],
                cauchy_minus.log_prob(z)[None, ...],
            ],
            0,
        )
        log_target = torch.logsumexp(catted, 0).sum(-1) - 2 * torch.tensor(2.0).log()
        # log_target = log_target[:, None] +

        return log_target  # + torch.tensor([8.]).log()

    def log_prob(self, z, x=None):
        catted = torch.cat(
            [
                self.cauchy.log_prob(z)[None, ...],
                self.cauchy_minus.log_prob(z)[None, ...],
            ],
            0,
        )
        log_target = (
            torch.logsumexp(catted, 0).sum(-1) - self.dim * torch.tensor(2.0).log()
        )

        # print(z.shape, log_target.shape)

        return log_target  # + torch.tensor([8.]).log()


class Funnel(Distribution):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.device = kwargs.get("device", "cpu")
        self.a = (kwargs.get("a", 1.0) * torch.ones(1)).to(self.device)
        self.b = kwargs.get("b", 0.5)
        self.dim = kwargs.get("dim", 16)
        self.distr1 = torch.distributions.Normal(torch.zeros(1).to(self.device), self.a)
        # self.distr2 = lambda z1: torch.distributions.MultivariateNormal(torch.zeros(self.dim-1), (2*self.b*z1).exp()*torch.eye(self.dim-1))
        # self.distr2 = lambda z1: -(z[...,1:]**2).sum(-1) * (-2*self.b*z1).exp() - np.log(self.dim) + 2*self.b*z1
        self.xlim = [-2, 10]
        self.ylim = [-30, 30]
        self.scale_2d_log_prob = 20  # 30.0

    def log_prob(self, z, x=None):
        logprob1 = self.distr1.log_prob(z[..., 0])
        z1 = z[..., 0]
        logprob2 = (
            -0.5 * (z[..., 1:] ** 2).sum(-1) * torch.exp(-2 * self.b * z1)
            - (self.dim - 1) * self.b * z1
        )
        return logprob1 + logprob2

    def log_prob_2d_slice(self, z, dim1=0, dim2=1):
        if dim1 == 0 or dim2 == 0:
            logprob1 = self.distr1.log_prob(z[..., 0])
            dim2 = dim2 if dim2 != 0 else dim1
            z1 = z[..., 0]
            # logprob2 = self.distr2(z[...,0])
            logprob2 = (
                -0.5 * (z[..., dim2] ** 2) * torch.exp(-2 * self.b * z1) - self.b * z1
            )
        # else:
        #     logprob2 = -(z[...,dim2]**2) * (-2*self.b*z1).exp() - np.log(self.dim) + 2*self.b*z1
        return logprob1 + logprob2

    def plot_2d_countour(self, ax):
        x = np.linspace(-15, 15, 100)
        y = np.linspace(-10, 10, 100)
        X, Y = np.meshgrid(x, y)
        inp = torch.from_numpy(np.stack([X, Y], -1))
        Z = self.log_prob(inp.reshape(-1, 2)).reshape(inp.shape[:-1])

        # levels = np.quantile(Z, np.linspace(0.9, 0.99, 5))
        ax.contour(
            X,
            Y,
            Z.exp(),
            # levels = levels,
            levels=3,
            alpha=1.0,
            cmap="inferno",
        )

    def plot_2d(self, fig=None, ax=None):
        if fig is None and ax is None:
            fig, ax = plt.subplots()

            xlim = [-2, 13]
            ylim = [-60, 60]
        else:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()

        x = np.linspace(*xlim, 100)
        y = np.linspace(*ylim, 100)
        xx, yy = np.meshgrid(x, y)
        z = torch.FloatTensor(np.stack([xx, yy], -1))
        vals = (self.log_prob_2d_slice(z) / self.scale_2d_log_prob).exp()
        if ax is not None:
            ax.imshow(
                vals.flip(0),
                extent=[*xlim, *ylim],
                cmap="Greens",
                alpha=0.5,
                aspect="auto",
            )
        else:
            plt.imshow(
                vals.flip(0),
                extent=[*xlim, *ylim],
                cmap="Greens",
                alpha=0.5,
                aspect="auto",
            )

        return fig, xlim, ylim


class HalfBanana(Distribution):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.Q = kwargs.get("Q", 0.01) * torch.ones(1)
        self.dim = kwargs.get("dim", 32)
        self.xlim = [-1, 9]
        self.ylim = [-2, 4]
        self.scale_2d_log_prob = 2.0
        # assert self.dim % 2 == 0, 'Dimension should be divisible by 2'

    def log_prob(self, z, x=None):
        # n = self.dim/2
        even = np.arange(0, self.dim, 2)
        odd = np.arange(1, self.dim, 2)

        ll = -((z[..., even] - z[..., odd] ** 2) ** 2) / self.Q - (z[..., odd] - 1) ** 2
        return ll.sum(-1)

    def log_prob_2d_slice(self, z, dim1=0, dim2=1):
        if dim1 % 2 == 0 and dim2 % 2 == 1:
            ll = (
                -((z[..., dim1] - z[..., dim2] ** 2) ** 2) / self.Q
                - (z[..., dim2] - 1) ** 2
            )
        return ll  # .sum(-1)


class Banana(Distribution):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.device = kwargs.get("device", "cpu")
        self.b = kwargs.get("b", 0.02)  # * torch.ones(1).to(self.device)
        self.sigma = kwargs.get("sigma", 100.0)  # * torch.ones(1).to(self.device)
        self.dim = kwargs.get("dim", 32)
        self.xlim = [-1, 5]
        self.ylim = [-2, 2]
        self.scale_2d_log_prob = 2.0
        # assert self.dim % 2 == 0, 'Dimension should be divisible by 2'

    def log_prob(self, z, x=None):
        # n = self.dim/2
        even = np.arange(0, self.dim, 2)
        odd = np.arange(1, self.dim, 2)

        ll = -0.5 * (
            z[..., odd] - self.b * z[..., even] ** 2 + (self.sigma**2) * self.b
        ) ** 2 - ((z[..., even]) ** 2) / (2 * self.sigma**2)
        return ll.sum(-1)

    def log_prob_2d_slice(self, z, dim1=0, dim2=1):
        if dim1 % 2 == 0 and dim2 % 2 == 1:
            ll = (
                -((z[..., dim1] - z[..., dim2] ** 2) ** 2) / self.Q
                - (z[..., dim1] - 1) ** 2
            )
        return ll  # .sum(-1)


class BayesianLogRegression(Distribution):
    tau: float
    # dataset : ClassificationDataset

    def __init__(self, dataset: ClassificationDataset, **kwargs):
        super().__init__(**kwargs)
        # self.dataset = dataset
        self.x_train = dataset.x_train
        self.y_train = dataset.y_train
        self.d = dataset.d
        self.n = dataset.n
        self.tau = kwargs.get("tau", 0.05)

    def log_prob(self, theta, x=None, y=None):
        if x is None:
            x = self.x_train
        if y is None:
            y = self.y_train

        max_val = 1e5

        prod = torch.clamp(
            torch.matmul(x, theta.transpose(0, 1)),
            min=-max_val,
        )
        # P = 1. / (1. + torch.exp(-prod))
        P = torch.sigmoid(prod)
        # mask = torch.isnan(P)
        # P = P[mask]
        # P[mask] = 1e-5
        ll = torch.matmul(
            y,
            torch.log(torch.clamp(P, min=1e-5)),
        ) + torch.matmul(1 - y, torch.log(torch.clamp(1 - P, min=1e-5)))

        ll = ll - self.tau / 2 * (theta**2).sum(-1)
        return ll

    # def grad_log_prob(self, theta, x= None, y = None):
    #     if x is None:
    #         x = self.x_train
    #     if y is None:
    #         y = self.y_train
    #     P = 1. / (1. + torch.exp(-torch.mm(x,theta)))
    #     return torch.matmul(torch.transpose(X, -2, -1), (y - P).view(self.n)) - self.tau * theta

    # def log_prob(self, theta, x= None, y = None):
    #     if x is None:
    #         x = self.x_train
    #     if y is None:
    #         y = self.y_train
    #     P = 1. / (1. + torch.exp(-torch.matmul(x,theta.transpose(0,1))))
    #     ll =  torch.matmul(y,torch.log(P)) + torch.matmul(1-y,torch.log(1-P)) - self.tau/2 * (theta**2).sum(-1)
    #     return ll


class BayesianLinearRegression(Distribution):
    def __init__(self, dataset: RegressionDataset, **kwargs):
        super().__init__(**kwargs)
        self.x_train = dataset.x_train
        self.y_train = dataset.y_train
        self.d = dataset.d
        self.n = dataset.n
        self.tau = kwargs.get("tau", 0.05)
        self.noise_scale = kwargs.get("noise_scale", 0.1)

    def log_prob(self, theta, x=None, y=None):
        if x is None:
            x = self.x_train
        if y is None:
            y = self.y_train

        # max_val = 1e5

        ll = -((y - torch.dot(x, theta)) ** 2) / (2 * self.noise_scale**2)

        # prod = torch.clamp(torch.matmul(x, theta.transpose(0,1)), min=-max_val)
        # #P = 1. / (1. + torch.exp(-prod))
        # P = torch.sigmoid(prod)
        # #print(P.max())
        # mask = torch.isnan(P)
        # #P = P[mask]
        # #P[mask] = 1e-5
        # ll =  torch.matmul(y, torch.log(torch.clamp(P, min=1e-5))) + torch.matmul(1-y, torch.log(torch.clamp(1 - P, min=1e-5)))

        ll = ll - self.tau / 2 * (theta**2).sum(-1)
        return ll


class GMM(nn.Module):
    def __init__(self):
        super().__init__()
        self.logits = nn.Parameter(self.logits_prior.sample((self.n_samples,)))
        self.log_vars = nn.Parameter(
            self.vars_prior.sample(
                (self.n_samples, self.n_components, self.dim),
            ).double(),
        )
        self.means = nn.Parameter(
            torch.normal(0.0, 2 * (0.5 * self.log_vars).exp()),
        )


class GaussianMixtureModel(Distribution, nn.Module):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.n_components: int = kwargs.get("n_components", 5)
        self.dim: int = kwargs.get("n_components", 5)
        self.n_samples: int = kwargs.get("n_samples", 1)
        self.stick_break = td.StickBreakingTransform()

        self.logits_prior = td.TransformedDistribution(
            td.Dirichlet(torch.ones(self.n_components, dtype=torch.double)),
            self.stick_break.inv,
        )

        self.vars_prior = td.TransformedDistribution(
            td.Gamma(1, 1),
            td.ComposeTransform(
                [td.ExpTransform().inv, td.AffineTransform(0, -1)],
            ),
        )

        # self.model = GMM(self.logits_prior, self.vars_prior, self.n_samples, self.n_components, self.dim)

        with torch.no_grad():
            self.logits = nn.Parameter(
                self.logits_prior.sample((self.n_samples,)),
            )
            self.log_vars = nn.Parameter(
                self.vars_prior.sample(
                    (self.n_samples, self.n_components, self.dim),
                ).double(),
            )
            self.means = nn.Parameter(
                torch.normal(0.0, 2 * (0.5 * self.log_vars).exp()),
            )

        self.data = self.sample_data(n_samples=1000)

    def forward(self, x):
        sigmas = (0.5 * self.log_vars).exp()
        betas = self.stick_break(self.logits)

        x = x[:, None, None, :]

        log_ps = td.Normal(self.means, sigmas).log_prob(x).sum(dim=-1)
        log_ps = torch.logsumexp(log_ps + betas.log(), dim=2)

        return log_ps

    def log_prior(self):
        p_logits = self.logits_prior.log_prob(self.logits)
        p_sigma = self.vars_prior.log_prob(self.log_vars).sum(dim=-1).sum(dim=-1)
        means_prior = td.Normal(0.0, 2 * (0.5 * self.log_vars).exp())
        p_means = means_prior.log_prob(self.means).sum(dim=-1).sum(dim=-1)

        return p_logits + p_sigma + p_means

    def sample_data(self, n_samples: int = 1000):
        with torch.no_grad():
            sigmas = (0.5 * self.log_vars[0]).exp()
            betas = self.stick_break(self.logits[0])
            mus = self.means[0]

            zs = rng.choice(self.n_components, p=betas.numpy(), size=n_samples)
            xs = torch.normal(mus[zs], sigmas[zs])

        return xs


class PhiFour(Distribution):
    """
    Borrowed from oficial implementation: https://github.com/marylou-gabrie/flonaco
    Original paper: https://arxiv.org/pdf/2105.12603.pdf
    """

    def __init__(
        self,
        a,
        b,
        dim_grid,
        dim_phys=1,
        beta=1,
        bc=("dirichlet", 0),
        tilt=None,
        dtype=torch.float32,
        device="cpu",
    ):
        """
        Class to handle operations around Stochastic Allen-Cahn model
        Args:
            a: coupling term coef
            b: local field coef
            dim_grid: grid size along one physical dimension
            dim_phys: number of dimensions of the physical grid
            beta: inverse temperature
            tilt: None or {"val":0.7, "lambda":1} - mean value + Lagrange param
        """
        super().__init__(device=device)
        self.device = device

        self.a = a
        self.b = b
        self.beta = beta
        self.dim_grid = dim_grid
        self.dim_phys = dim_phys
        self.sum_dims = tuple(i + 1 for i in range(dim_phys))

        self.bc = bc
        self.tilt = tilt

    def init_field(self, n_or_values):
        if isinstance(n_or_values, int):
            x = torch.rand((n_or_values,) + (self.dim_grid,) * self.dim_phys)
            x = x * 2 - 1
        else:
            x = n_or_values
        return x

    def V(self, x):
        coef = self.a * self.dim_grid
        V = ((1 - x**2) ** 2 / 4 + self.b * x).sum(self.sum_dims) / coef
        if self.tilt is not None:
            tilt = (self.tilt["val"] - x.mean(self.sum_dims)) ** 2
            tilt = self.tilt["lambda"] * tilt / (4 * self.dim_grid)
            V += tilt
        return V

    def log_prob(self, x):
        # Does not include the temperature! need to be explicitely added in Gibbs factor

        if self.bc[0] == "dirichlet":
            x_ = F.pad(
                input=x,
                pad=(1,) * (2 * self.dim_phys),
                mode="constant",
                value=self.bc[1],
            )
        else:
            raise NotImplementedError("Only dirichlet BC implemeted")

        grad_term = ((x_[:, 1:, ...] - x_[:, :-1, ...]) ** 2 / 2).sum(
            self.sum_dims,
        )
        if self.dim_phys == 2:
            grad_term += ((x_[:, :, 1:] - x_[:, :, :-1]) ** 2 / 2).sum(
                self.sum_dims,
            )

        coef = self.a * self.dim_grid
        return -(grad_term * coef + self.V(x))

    def U_coupling_per_site(self, x):
        """
        return the (\nabla phi) ** 2 to be used in direct computation
        """
        assert self.dim_phys == 1
        if self.bc[0] == "dirichlet":
            x_ = F.pad(
                input=x,
                pad=(1,) * (2 * self.dim_phys),
                mode="constant",
                value=self.bc[1],
            )
        else:
            raise NotImplementedError("Only dirichlet BC implemeted")

        return ((x_[:, 1:] - x_[:, :-1]) ** 2 / 2) * self.a * self.dim_grid
